//*****************************************************************************
//* =====================
//*  HCHCSpring
//* =====================
//*  
//* (Description)
//*    2HDM: e+e- --> H+H- generator
//*
//* (Update Record)
//*    2017/01/19  K.Fujii	Original version.
//*
//*****************************************************************************

#include "JSFSteer.h"
#include "HCHCSpring.h"
#include "GENNumCon.h"

#include <sstream>
#include <iomanip>

//#define __NODECAY__
//#define __DEBUG__
//#define __ZEROWIDTH__
//#define __PAHSESPACE__
#ifdef __NODECAY__
#ifndef __ZEROWIDTH__
#define __ZEROWIDTH__
#endif
#endif
//#define TEMP_H

//*----------------------------------------------------------------------
//*     Numerical and Natural Constants
//*----------------------------------------------------------------------
#include "GENNumCon.h"

ClassImp(HCHCSpring)
ClassImp(HCHCSpringBuf)
ClassImp(HCHCBases)
ClassImp(HCBoson)
ClassImp(TopQuark)

//-----------------------------------------------------------------------------
// ==============================
//  class HCHCSpring
// ==============================
//_____________________________________________________________________________
// --------------------------
//  C-tor
// --------------------------
HCHCSpring::HCHCSpring(const char      *name,
                       const char      *title,
                             HCHCBases *bases)
            : JSFSpring(name, title, bases)
{
  fEventBuf = new HCHCSpringBuf("HCHCSpringBuf",
                                "HCHCSpring event buffer",
                                this);
  if (!bases) { 
    SetBases(new HCHCBases);
  }
}

//_____________________________________________________________________________
// --------------------------
//  D-tor
// --------------------------
HCHCSpring::~HCHCSpring()
{
  delete fEventBuf;
  delete GetBases();
}

//_____________________________________________________________________________
// --------------------------
//  Initialize
// --------------------------
Bool_t HCHCSpring::Initialize()
{
  JSFSpring::Initialize();

  if (fFile->IsWritable()) {
    HCHCBases *bs = static_cast<HCHCBases *>(GetBases());
    TDirectory *last = gDirectory;
    fFile->cd("/conf");
    TList *dlist = gDirectory->GetListOfKeys();
    if (!dlist->FindObject("init")) gDirectory->mkdir("init");
    fFile->cd("/conf/init");
    bs->Write();
    last->cd(); 
    cerr << ">>>>>> HCHCBases written to file" << endl;
  }

  return kTRUE;
}

//-----------------------------------------------------------------------------
// ==============================
//  class HCHCSpringBuf
// ==============================
//_____________________________________________________________________________
// --------------------------
//  SetPartons
// --------------------------
//                                      _
//                                 / 6 b
//                                /
//                               /         _
//                           _  /\7 W-/ 10 f(up)
//       e- \              3 t /  \  /
//           \                /    \/
//            \         0 H- /\     \
//             \            /  \     \
//              \          /    \     \ 11 f(down)
//               \________/       2 b
//               /        \         _
//              /          \    / 4 b /    _
//             /            \  /     /  13 f(down)
//            /         1 H+ \/     /  
//           /                \    /\
//      e+  /              5 t \  /  \
//                              \/9 W+\ 12 f(up)
//                               \
//                                \
//                                 \ 8 b(down)
//                               

Bool_t HCHCSpringBuf::SetPartons()
{
  TClonesArray &partons = *fPartons;
  HCHCBases  *bases   = (HCHCBases*)((HCHCSpring*)Module())->GetBases();

  // ----------------------------------------------
  //  Set 4-momenta of final state particles in CM
  // ----------------------------------------------
  const Int_t kNparton = 14;
  ANL4DVector pv[kNparton];
  pv[ 2] = bases->fP[0];    // b   from H-
  pv[ 6] = bases->fP[1];    // bb  from tb
  pv[10] = bases->fP[2];    // fub from W-
  pv[11] = bases->fP[3];    // fd  from W-

  pv[ 7] = pv[10] + pv[11]; // W-
  pv[ 3] = pv[ 6] + pv[ 7]; // tb

  pv[ 4] = bases->fP[4];    // bb  from H+
  pv[ 8] = bases->fP[5];    // b   from t
  pv[12] = bases->fP[6];    // fu  from W+
  pv[13] = bases->fP[7];    // fdb from W+

  pv[ 9] = pv[12] + pv[13]; // W+
  pv[ 5] = pv[ 8] + pv[ 9]; // t

  pv[ 0] = pv[ 2] + pv[ 3]; // H-
  pv[ 1] = pv[ 4] + pv[ 5]; // H+

  // ----------------------------------------------
  //  Boost them to lab. frame
  // ----------------------------------------------

  fZBoost       = bases->GetZBoost();
  fEcmIP        = bases->GetEcmIP();

  fQ2HH         = fEcmIP*fEcmIP;
  fCosTheta     = bases->GetCosTheta();
  fPhi          = bases->GetPhi();

  fQ2HM         = bases->GetQ2HM();
  fCosThetaT1   = bases->GetCosThetaT1();
  fPhiT1        = bases->GetPhiT1();

  fQ2T1         = bases->GetQ2T1();
  fCosThetaWM   = bases->GetCosThetaWM();
  fPhiWM        = bases->GetPhiWM();

  fQ2WM         = bases->GetQ2WM();
  fCosThetaF1   = bases->GetCosThetaF1();
  fPhiF1        = bases->GetPhiF1();

  fQ2HP         = bases->GetQ2HP();
  fCosThetaT2   = bases->GetCosThetaT2();
  fPhiT2        = bases->GetPhiT2();

  fQ2T2         = bases->GetQ2T2();
  fCosThetaWP   = bases->GetCosThetaWP();
  fPhiWP        = bases->GetPhiWP();

  fQ2WP         = bases->GetQ2WP();
  fCosThetaF2   = bases->GetCosThetaF2();
  fPhiF2        = bases->GetPhiF2();

  Double_t elab = TMath::Sqrt(fEcmIP*fEcmIP + fZBoost*fZBoost);
  TVector3 boostv(0.,0.,fZBoost/elab);
  for (Int_t i=0; i<kNparton; i++) pv[i].Boost(boostv);

  // ----------------------------------------------
  //  Set final state parton infomation
  // ----------------------------------------------

  fNparton = kNparton;

  static TVector **qp = 0;
  if (!qp) {
    qp = new TVector* [kNparton];
    for (Int_t i=0; i<kNparton; i++) qp[i] = new TVector(4);
  }
  for (Int_t i=0; i<kNparton; i++) {
    TVector &q = *qp[i];
    q(0) = pv[i].E ();
    q(1) = pv[i].Px();
    q(2) = pv[i].Py();
    q(3) = pv[i].Pz();
  }
  Int_t    idx     = bases->fHCBosonPtr->GetPID(); // PDG code for H+
  Int_t    idb     =  5;                           // PDG code for b
  Int_t    idt     =  6;                           // PDG code for b
  Int_t    idw     = 24;                           // PDG code for W+

  Double_t mb      = kMass[1][1][2];

  Int_t    helbhm  = bases->fHelFinal[0];
  Int_t    helbt1  = bases->fHelFinal[1];
  Int_t    islevhm = 101;                          // shower level
  Int_t    icfhm   = 1;                            // color flux id
  Int_t    islevt1 = 102;                          // shower level

  Int_t    id1     = bases->f1Ptr->GetPID();       // PDG code for 1st daughter
  Int_t    id2     = bases->f2Ptr->GetPID();       // PDG code for 2nd daughter
  Double_t m11     = bases->f1Ptr->GetMass();      // 1st daughter mass
  Double_t m12     = bases->f2Ptr->GetMass();      // 2nd daughter mass
  Int_t    hel1    = bases->fHelFinal[2];          // 1st daughter helicity
  Int_t    hel2    = bases->fHelFinal[3];          // 2nd daughter helicity
  Double_t colorm  = bases->f1Ptr->GetColor();     // color factor
  Double_t chg1    = bases->f1Ptr->GetCharge();    // 1st daughter charge
  Double_t chg2    = bases->f2Ptr->GetCharge();    // 2nd daughter charge
  Int_t    islevm  = colorm > 1. ? 201 : 0;        // shower level
  Int_t    icfm    = 2;                            // color flux id

  Int_t    helbhp  = bases->fHelFinal[4];
  Int_t    helbt2  = bases->fHelFinal[5];
  Int_t    islevhp = 301;                          // shower level
  Int_t    icfhp   = 3;                            // color flux id
  Int_t    islevt2 = 302;                          // shower level

  Int_t    id4     = bases->f4Ptr->GetPID();       // PDG code for 1st daughter
  Int_t    id5     = bases->f5Ptr->GetPID();       // PDG code for 2nd daughter
  Double_t m13     = bases->f4Ptr->GetMass();      // 1st daughter mass
  Double_t m14     = bases->f5Ptr->GetMass();      // 2nd daughter mass
  Int_t    hel4    = bases->fHelFinal[6];          // 4th daughter helicity
  Int_t    hel5    = bases->fHelFinal[7];          // 5th daughter helicity
  Double_t colorp  = bases->f4Ptr->GetColor();     // color factor
  Double_t chg4    = bases->f4Ptr->GetCharge();    // 1st daughter charge
  Double_t chg5    = bases->f5Ptr->GetCharge();    // 2nd daughter charge
  Int_t    islevp  = colorp > 1. ? 401 : 0;        // shower level
  Int_t    icfp    = 4;                            // color flux id

  Double_t rq2hm   = pv[0].Mag();
  Double_t rq2hp   = pv[1].Mag();
  Double_t rq2t1   = pv[3].Mag();
  Double_t rq2t2   = pv[5].Mag();
  Double_t rq2wm   = pv[7].Mag();
  Double_t rq2wp   = pv[9].Mag();

  Double_t mass    = bases->GetMass();

  //                               No. PID   Mass   Charge   pv    Nd 1st Mom     hel    col  shower
  new (partons[ 0]) JSFSpringParton( 1,-idx, mass ,   -1., *qp[ 0], 2,  3,  0,      0,     0,       0);
  new (partons[ 1]) JSFSpringParton( 2,+idx, mass ,   +1., *qp[ 1], 2,  5,  0,      0,     0,       0);
  new (partons[ 2]) JSFSpringParton( 3, idb, mb   , -1./3, *qp[ 2], 0,  0,  1, helbhm, icfhm, islevhm);
  new (partons[ 3]) JSFSpringParton( 4,-idt, rq2t1, -2./3, *qp[ 3], 2,  7,  1,      0,    10, islevhm);
  new (partons[ 4]) JSFSpringParton( 5,-idb, mb   , +1./3, *qp[ 4], 0,  0,  2, helbhp, icfhp, islevhp);
  new (partons[ 5]) JSFSpringParton( 6, idt, rq2t2, +2./3, *qp[ 5], 2,  9,  2,      0,    20, islevhp);

  new (partons[ 6]) JSFSpringParton( 7,-idb, mb   , +1./3, *qp[ 6], 0,  0,  4, helbt1, icfhm, islevt1);
  new (partons[ 7]) JSFSpringParton( 8,-idw, rq2wm,   -1., *qp[ 7], 2, 10,  4,      0,    30, islevt1);
  new (partons[ 8]) JSFSpringParton( 9, idb, mb   , -1./3, *qp[ 8], 0,  0,  6, helbt2, icfhp, islevt2);
  new (partons[ 9]) JSFSpringParton(10, idw, rq2wp,   +1., *qp[ 9], 2, 12,  6,      0,    40, islevt2);

  new (partons[10]) JSFSpringParton(11,-id1, m11  , -chg1, *qp[10], 0,  8,  4, hel1  , icfm , islevm );
  new (partons[11]) JSFSpringParton(12,-id2, m12  , -chg2, *qp[11], 0,  8,  4, hel2  , icfm , islevm );
  new (partons[12]) JSFSpringParton(13, id4, m13  ,  chg4, *qp[12], 0, 10,  6, hel4  , icfp , islevp );
  new (partons[13]) JSFSpringParton(14, id5, m14  ,  chg5, *qp[13], 0, 10,  6, hel5  , icfp , islevp );

  return kTRUE ;
}

//-----------------------------------------------------------------------------
// ==============================
//  class HCHCBases
// ==============================
//_____________________________________________________________________________
// --------------------------
//  C-tor
// --------------------------
HCHCBases::HCHCBases(const char *name, const char *title)
         : JSFBases   (name, title), 
           fMass      ( 300.),
           fTopMass   ( 175.),
           fTanBeta   ( 2.),
           fEcmInit   ( 500.),
           fISR       ( 1),
           fBeamStr   ( 1),
           fBeamWidth (0.002),
           fPole      (0.),
           fWmModesLo ( 1),
           fWmModesHi (12),
           fWpModesLo ( 1),
           fWpModesHi (12),
           fHCBosonPtr( 0),
           fWMBosonPtr( 0),
           fWPBosonPtr( 0),
           fZBosonPtr ( 0),
           fPhotonPtr ( 0),
           fGluonPtr  ( 0),
           fZBoost    (0.),
           fEcmIP     (fEcmInit),
           fQ2HH      (0.),
           fQ2HM      (0.),
           fQ2HP      (0.),
           fQ2T1      (0.),
           fQ2T2      (0.),
           fT1Ptr     (0),
           fWMModePtr (0),
           f1Ptr      (0),
           f2Ptr      (0),
           fQ2WM      (0.),
           fT2Ptr     (0),
           fWPModePtr (0),
           f4Ptr      (0),
           f5Ptr      (0),
           fQ2WP      (0.),
           fCosTheta  (0.),
           fPhi       (0.),
           fCosThetaT1(0.),
           fPhiT1     (0.),
           fXQ2HM     (0.),
           fCosThetaWM(0.),
           fPhiWM     (0.),
           fXQ2T1     (0.),
           fCosThetaF1(0.),
           fPhiF1     (0.),
           fXQ2WM     (0.),
           fCosThetaT2(0.),
           fPhiT2     (0.),
           fXQ2HP     (0.),
           fCosThetaWP(0.),
           fPhiWP     (0.),
           fXQ2T2     (0.),
           fCosThetaF2(0.),
           fPhiF2     (0.),
           fXQ2WP     (0.),
           fR_BW_m    (0),
           fR_BW_p    (0),
           fR_BS_m    (0),
           fR_BS_p    (0),
           fR_ISR_var (0),
           fR_ISR_side(0)
{
  //  Constructor of bases.  Default parameter should be initialized here
  //
  // --------------------------------------------
  //  Get parameters from jsf.conf, if specified
  // --------------------------------------------

  cout << "Init hchcbases " << endl;
  
  using namespace std;
  stringstream ins(gJSF->Env()->GetValue("HCHCBases.CosthRange","-1.0 1.0"));
  ins >> fXL[0] >> fXU[0];

  ins.clear();
  ins.str(gJSF->Env()->GetValue("HCHCBases.PhiOverPiRange","0.0 2.0"));
  ins >> fXL[1] >> fXU[1];

  ins.clear();
  ins.str(gJSF->Env()->GetValue("HCHCBases.CosthT1Range","-1.0 1.0"));
  ins >> fXL[2] >> fXU[2];

  ins.clear();
  ins.str(gJSF->Env()->GetValue("HCHCBases.PhiT1OverPiRange","0.0 2.0"));
  ins >> fXL[3] >> fXU[3];

  ins.clear();
  ins.str(gJSF->Env()->GetValue("HCHCBases.CosthWmRange","-1.0 1.0"));
  ins >> fXL[4] >> fXU[4];

  ins.clear();
  ins.str(gJSF->Env()->GetValue("HCHCBases.PhiWmOverPiRange","0.0 2.0"));
  ins >> fXL[5] >> fXU[5];

  ins.clear();
  ins.str(gJSF->Env()->GetValue("HCHCBases.CosthFmRange","-1.0 1.0"));
  ins >> fXL[6] >> fXU[6];

  ins.clear();
  ins.str(gJSF->Env()->GetValue("HCHCBases.PhiFmOverPiRange","0.0 2.0"));
  ins >> fXL[7] >> fXU[7];


  ins.clear();
  ins.str(gJSF->Env()->GetValue("HCHCBases.CosthT2Range","-1.0 1.0"));
  ins >> fXL[8] >> fXU[8];

  ins.clear();
  ins.str(gJSF->Env()->GetValue("HCHCBases.PhiT2OverPiRange","0.0 2.0"));
  ins >> fXL[9] >> fXU[9];

  ins.clear();
  ins.str(gJSF->Env()->GetValue("HCHCBases.CosthWpRange","-1.0 1.0"));
  ins >> fXL[10] >> fXU[10];

  ins.clear();
  ins.str(gJSF->Env()->GetValue("HCHCBases.PhiWpOverPiRange","0.0 2.0"));
  ins >> fXL[11] >> fXU[11];

  ins.clear();
  ins.str(gJSF->Env()->GetValue("HCHCBases.CosthFpRange","-1.0 1.0"));
  ins >> fXL[12] >> fXU[12];

  ins.clear();
  ins.str(gJSF->Env()->GetValue("HCHCBases.PhiFpOverPiRange","0.0 2.0"));
  ins >> fXL[13] >> fXU[13];

  ins.clear();
  ins.str(gJSF->Env()->GetValue("HCHCBases.Mass","300.")); 	 // M_H+ [GeV]
  ins >> fMass;

  ins.clear();
  ins.str(gJSF->Env()->GetValue("HCHCBases.TopMass","175.")); 	 // M_top [GeV]
  ins >> fTopMass;

  ins.clear();
  ins.str(gJSF->Env()->GetValue("HCHCBases.TanBeta","2.")); 	 // tan(beta)
  ins >> fTanBeta;

  ins.clear();
  ins.str(gJSF->Env()->GetValue("HCHCBases.Ecm","500."));       // E_cm (1TeV)
  ins >> fEcmInit;

  ins.clear();
  ins.str(gJSF->Env()->GetValue("HCHCBases.Beamstrahlung","1")); // BmStr (on)
  ins >> fBeamStr;

  ins.clear();
  ins.str(gJSF->Env()->GetValue("HCHCBases.BeamWidth","0.002")); // Beam width
  ins >> fBeamWidth;

  ins.clear();
  ins.str(gJSF->Env()->GetValue("HCHCBases.Bremsstrahlung","1"));// ISR (on)
  ins >> fISR;

  ins.clear();
  ins.str(gJSF->Env()->GetValue("HCHCBases.Pole","0."));         // electron polarization
  ins >> fPole;

  ins.clear();
  ins.str(gJSF->Env()->GetValue("HCHCBases.WmModesLo","1"));     // W- decay mode lo
  ins >> fWmModesLo;

  ins.clear();
  ins.str(gJSF->Env()->GetValue("HCHCBases.WmModesHi","12"));    // W- decay mode hi
  ins >> fWmModesHi;

  ins.clear();
  ins.str(gJSF->Env()->GetValue("HCHCBases.WpModesLo","1"));     // W+ decay mode lo
  ins >> fWpModesLo;

  ins.clear();
  ins.str(gJSF->Env()->GetValue("HCHCBases.WpModesHi","12"));    // W+ decay mode hi
  ins >> fWpModesHi;

  // --------------------------------------------
  //  Registor random numbers
  // --------------------------------------------
  //--
  //  Initial state helicity combination
  //--
  DefineVariable(fHelCombInitial, 0., 1., 0, 1);
  DefineVariable(fHelCombFinal  , 0., 1., 0, 1);
  DefineVariable(fWMDecayMode   , 0., 1., 0, 1);
  DefineVariable(fWPDecayMode   , 0., 1., 0, 1);
  DefineVariable(fXQ2HM         , 0., 1., 0, 1);
  DefineVariable(fXQ2HP         , 0., 1., 0, 1);
  DefineVariable(fXQ2T1         , 0., 1., 0, 1);
  DefineVariable(fXQ2T2         , 0., 1., 0, 1);
  DefineVariable(fXQ2WM         , 0., 1., 0, 1);
  DefineVariable(fXQ2WP         , 0., 1., 0, 1);
  //--
  //  cos(theta) and phi
  //--
  fXL[ 1] = fXL[ 1]*TMath::Pi();
  fXU[ 1] = fXU[ 1]*TMath::Pi();
  fXL[ 3] = fXL[ 3]*TMath::Pi();
  fXU[ 3] = fXU[ 3]*TMath::Pi();
  fXL[ 5] = fXL[ 5]*TMath::Pi();
  fXU[ 5] = fXU[ 5]*TMath::Pi();
  fXL[ 7] = fXL[ 7]*TMath::Pi();
  fXU[ 7] = fXU[ 7]*TMath::Pi();
  fXL[ 9] = fXL[ 9]*TMath::Pi();
  fXU[ 9] = fXU[ 9]*TMath::Pi();
  fXL[11] = fXL[11]*TMath::Pi();
  fXU[11] = fXU[11]*TMath::Pi();
  fXL[13] = fXL[13]*TMath::Pi();
  fXU[13] = fXU[13]*TMath::Pi();

  DefineVariable(fCosTheta  , fXL[ 0], fXU[ 0], 1, 1);
  DefineVariable(fPhi       , fXL[ 1], fXU[ 1], 0, 0);
  DefineVariable(fCosThetaT1, fXL[ 2], fXU[ 2], 0, 0);
  DefineVariable(fPhiT1     , fXL[ 3], fXU[ 3], 0, 0);
  DefineVariable(fCosThetaWM, fXL[ 4], fXU[ 4], 0, 1);
  DefineVariable(fPhiWM     , fXL[ 5], fXU[ 5], 0, 1);
  DefineVariable(fCosThetaF1, fXL[ 6], fXU[ 6], 0, 1);
  DefineVariable(fPhiF1     , fXL[ 7], fXU[ 7], 0, 1);

  DefineVariable(fCosThetaT2, fXL[ 8], fXU[ 8], 0, 0);
  DefineVariable(fPhiT2     , fXL[ 9], fXU[ 9], 0, 0);
  DefineVariable(fCosThetaWP, fXL[10], fXU[10], 0, 1);
  DefineVariable(fPhiWP     , fXL[11], fXU[11], 0, 1);
  DefineVariable(fCosThetaF2, fXL[12], fXU[12], 0, 1);
  DefineVariable(fPhiF2     , fXL[13], fXU[13], 0, 1);

  //--
  //  Beamstrahlung and natural Ebeam spread
  //--
  if (fBeamStr == 1) {
    DefineVariable(fR_BW_m, 0., 1., 0, 1);
    DefineVariable(fR_BW_p, 0., 1., 0, 1);
    DefineVariable(fR_BS_m, 0., 1., 0, 1);
    DefineVariable(fR_BS_p, 0., 1., 0, 1);
  }
 
  //--
  //  ISR
  //--
  if (fISR==1) {
    DefineVariable(fR_ISR_var,  0., 1., 0, 1);
    DefineVariable(fR_ISR_side, 0., 1., 0, 1);
  }

  // --------------------------------------------
  //  Set Bases integration parameters
  // --------------------------------------------
  //SetNoOfSample(100000);
  SetNoOfSample(10000);

  SetTuneValue (1.5);
  SetIteration1(0.05, 20);
  SetIteration2(0.05,100);
  //SetIteration1(0.05, 10);
  //SetIteration2(0.05, 20);

}
// --------------------------
//  D-tor
// --------------------------
HCHCBases::~HCHCBases()
{
  delete fHCBosonPtr;
  delete fT1Ptr;
  delete fT2Ptr;
  delete fWMBosonPtr;
  delete fWPBosonPtr;
  delete fZBosonPtr;
  delete fPhotonPtr;
  delete fGluonPtr;
}

//_____________________________________________________________________________
// --------------------------
//  Func
// --------------------------
Double_t HCHCBases::Func()
{
  //  Bases Integrand
  //

  Double_t bsWeight = 1.; // Jacobian factor
  Double_t eminus;        // E_e- after ISR and beamstrahlung
  Double_t eplus;         // E_e+ after ISR and beamstrahlung

  // --------------------------------------------
  //  Beamstrahlung
  // --------------------------------------------
  if (fBeamStr == 1) {
    bsWeight = fBM->GetWeight(fR_BW_m, fR_BW_p,
                              fR_BS_m, fR_BS_p,
                              eminus, eplus);
#if 0
    eplus   *= fBM->GetNominalEnergy();
    eminus  *= fBM->GetNominalEnergy();
#else
    eplus   *= fEcmInit/2.;
    eminus  *= fEcmInit/2.;
#endif
    fEcmIP   = 2.*TMath::Sqrt(eplus*eminus);     // ignore electron mass 
  } else {
    fEcmIP   = fEcmInit;      
    eplus    = fEcmInit/2.;
    eminus   = eplus;
  }
   
  // --------------------------------------------
  //  Initial State Radiation
  // --------------------------------------------

  static const Double_t kFisr = 2. * kAlpha0 / TMath::Pi();

  Double_t beta_isr = kFisr * (2.*(TMath::Log(fEcmIP/kM_e) - 1.));

  if (fISR == 1) {
    Double_t xisr    = TMath::Power(fR_ISR_var,1./beta_isr); // Ephoton / Ebeam 
             fEcmIP *= TMath::Sqrt(1. - xisr);                // reduced Ecm
    
    if (fR_ISR_side > 0.5) {
      eplus  *= 1. - xisr;
    } else {
      eminus *= 1. - xisr;
    }
  } // if Bremsstrahlung


  fZBoost = eminus - eplus; // P_z of the cm system after ISR and beamstrahlung

  // --------------------------------------------
  //  Select final state
  // --------------------------------------------
  Double_t weight = 1;

  Double_t mb  = kMass[1][1][2];
  Double_t m3  = mb;
  Double_t m5  = mb;
  Double_t m7  = mb;
  Double_t m9  = mb;

  fWMModePtr = fWMBosonPtr->PickMode(fWMDecayMode, weight, fWMMode);
#ifndef __NODECAY__
  bsWeight *= weight;
#endif
  f1Ptr = static_cast<GENPDTEntry *>(fWMModePtr->At(0));
  f2Ptr = static_cast<GENPDTEntry *>(fWMModePtr->At(1));
  Double_t m11  = f1Ptr->GetMass();
  Double_t m12  = f2Ptr->GetMass();

  fWPModePtr = fWPBosonPtr->PickMode(fWPDecayMode, weight, fWPMode);
#ifndef __NODECAY__
  bsWeight *= weight;
#endif
  f4Ptr = static_cast<GENPDTEntry *>(fWPModePtr->At(0));
  f5Ptr = static_cast<GENPDTEntry *>(fWPModePtr->At(1));
  Double_t m13  = f4Ptr->GetMass();
  Double_t m14  = f5Ptr->GetMass();

  // --------------------------------------------
  //  Check if reduced Ecm enough for prod.
  // --------------------------------------------
  if (fEcmIP < m3 + m5 + m7 + m9 + m11 + m12 + m13 + m14) {
    return 0.;
  }

  // --------------------------------------------
  //  Select helicity combination
  // --------------------------------------------
  //  Notice that spin average for e- is taken
  //  care of here
  SelectHelicities(weight);
  bsWeight *= weight;

  // --------------------------------------------
  //  Decide Q^2 of internal lines
  // --------------------------------------------
  fQ2HH = fEcmIP*fEcmIP;

  Double_t rs   = fEcmIP;
  Double_t qmin = m3 + m7 + m11 + m12;
  Double_t qmax = rs - (m5 + m9 + m13 + m14);
#ifndef __ZEROWIDTH__
  fQ2HM  = fHCBosonPtr->GetQ2BW(qmin, qmax, fXQ2HM, weight);
#else
  fQ2HM  = TMath::Power(fHCBosonPtr->GetMass(),2);
  weight = kPi*fHCBosonPtr->GetMass()*fHCBosonPtr->GetWidth();
#ifdef __NODECAY__
  weight = 1.;
#endif
#endif
  Double_t qx1 = TMath::Sqrt(fQ2HM);
  bsWeight *= weight;

#ifndef __ZEROWIDTH__
  qmin  = m5 + m9 + m13 + m14;
  qmax  = rs - qx1;
  fQ2HP = fHCBosonPtr->GetQ2BW(qmin, qmax, fXQ2HP, weight);
#else
  fQ2HP  = TMath::Power(fHCBosonPtr->GetMass(),2);
  weight = kPi*fHCBosonPtr->GetMass()*fHCBosonPtr->GetWidth();
#ifdef __NODECAY__
  weight = 1.;
#endif
#endif
  Double_t qx2 = TMath::Sqrt(fQ2HP);
  bsWeight *= weight;

#ifndef __ZEROWIDTH__
  qmin  = m7 + m11 + m12;
  qmax  = qx1 - m3;
  fQ2T1 = fT1Ptr->GetQ2BW(qmin, qmax, fXQ2T1, weight);
#else
  fQ2T1  = TMath::Power(fT1Ptr->GetMass(),2);
  weight = kPi*fT1Ptr->GetMass()*fT1Ptr->GetWidth();
#ifdef __NODECAY__
  weight = 1.;
#endif
#endif
  Double_t qt1 = TMath::Sqrt(fQ2T1);
  bsWeight *= weight;

#ifndef __ZEROWIDTH__
  qmin  = m9 + m13 + m14;
  qmax  = qx2 - m5;
  fQ2T2 = fT2Ptr->GetQ2BW(qmin, qmax, fXQ2T2, weight);
#else
  fQ2T2  = TMath::Power(fT2Ptr->GetMass(),2);
  weight = kPi*fT2Ptr->GetMass()*fT2Ptr->GetWidth();
#ifdef __NODECAY__
  weight = 1.;
#endif
#endif
  Double_t qt2 = TMath::Sqrt(fQ2T2);
  bsWeight *= weight;

#ifndef __ZEROWIDTH__
  qmin  = m11 + m12;
  qmax  = qt1 - m7;
  fQ2WM = fWMBosonPtr->GetQ2BW(qmin, qmax, fXQ2WM, weight);
#else
  fQ2WM  = TMath::Power(fWMBosonPtr->GetMass(),2);
  weight = kPi*fWMBosonPtr->GetMass()*fWMBosonPtr->GetWidth();
#ifdef __NODECAY__
  weight = 1.;
#endif
#endif
  bsWeight *= weight;

#ifndef __ZEROWIDTH__
  qmin  = m13 + m14;
  qmax  = qt2 - m9;
  fQ2WP = fWPBosonPtr->GetQ2BW(qmin, qmax, fXQ2WP, weight);
#else
  fQ2WP  = TMath::Power(fWPBosonPtr->GetMass(),2);
  weight = kPi*fWPBosonPtr->GetMass()*fWPBosonPtr->GetWidth();
#ifdef __NODECAY__
  weight = 1.;
#endif
#endif
  bsWeight *= weight;

  // --------------------------------------------
  //  Calcuate differential cross section
  // --------------------------------------------

  GENBranch w1branch(fQ2WM, fCosThetaF1, fPhiF1, m11*m11  , m12*m12);
  GENBranch t1branch(fQ2T1, fCosThetaWM, fPhiWM, &w1branch, m7*m7  );
  GENBranch x1branch(fQ2HM, fCosThetaT1, fPhiT1, &t1branch, m3*m3  );

  GENBranch w2branch(fQ2WP, fCosThetaF2, fPhiF2, m13*m13  , m14*m14);
  GENBranch t2branch(fQ2T2, fCosThetaWP, fPhiWP, &w2branch, m9*m9  );
  GENBranch x2branch(fQ2HP, fCosThetaT2, fPhiT2, &t2branch, m5*m5  );

  GENBranch cmbranch(fQ2HH, fCosTheta, fPhi, &x1branch, &x2branch);

  Double_t sigma = DSigmaDX(cmbranch);

  // --------------------------------------------
  //  Fill plots
  // --------------------------------------------

#ifdef TEMP_H
  H1Fill("h02", fCosTheta   , (bsWeight*sigma));
#endif
  Xh_fill( 1, fEcmIP            , (bsWeight*sigma));
  Xh_fill( 2, fCosTheta         , (bsWeight*sigma));
  Xh_fill( 3, fPhi              , (bsWeight*sigma));
  Xh_fill( 4, TMath::Sqrt(fQ2HM), (bsWeight*sigma));
  Xh_fill( 5, fCosThetaT1       , (bsWeight*sigma));
  Xh_fill( 6, fPhiT1            , (bsWeight*sigma));

  Xh_fill( 7, TMath::Sqrt(fQ2T1), (bsWeight*sigma));
  Xh_fill( 8, fCosThetaWM       , (bsWeight*sigma));
  Xh_fill( 9, fPhiWM            , (bsWeight*sigma));

  Xh_fill(10, TMath::Sqrt(fQ2WM), (bsWeight*sigma));
  Xh_fill(11, fCosThetaF1       , (bsWeight*sigma));
  Xh_fill(12, fPhiF1            , (bsWeight*sigma));

  Xh_fill(13, TMath::Sqrt(fQ2HP), (bsWeight*sigma));
  Xh_fill(14, fCosThetaT2       , (bsWeight*sigma));
  Xh_fill(15, fPhiT2            , (bsWeight*sigma));

  Xh_fill(16, TMath::Sqrt(fQ2T2), (bsWeight*sigma));
  Xh_fill(17, fCosThetaWP       , (bsWeight*sigma));
  Xh_fill(18, fPhiWP            , (bsWeight*sigma));

  Xh_fill(19, TMath::Sqrt(fQ2WP), (bsWeight*sigma));
  Xh_fill(20, fCosThetaF2       , (bsWeight*sigma));
  Xh_fill(21, fPhiF2            , (bsWeight*sigma));

  Xh_fill(22, (Double_t)fJCombI , (bsWeight*sigma));
  Xh_fill(23, (Double_t)fJCombF , (bsWeight*sigma));
  Xh_fill(24, (Double_t)fWMMode , (bsWeight*sigma));
  Xh_fill(25, (Double_t)fWPMode , (bsWeight*sigma));

  return (bsWeight * sigma);
}

//_____________________________________________________________________________
// --------------------------
//  DSigmaDX
// --------------------------
Double_t HCHCBases::DSigmaDX(GENBranch &cmbranch)
{
  // --------------------------------------------
  //  Phase space
  // --------------------------------------------
  Double_t rs = TMath::Sqrt(cmbranch.GetQ2());
  ANL4DVector qcm(rs, 0., 0., 0.);
  GENFrame    cmframe;

  // CM -> X- X+
  Double_t q2x1  = cmbranch.GetM12();
  Double_t q2x2  = cmbranch.GetM22();
  Double_t cosx1 = cmbranch.GetCosTheta();
  Double_t phix1 = cmbranch.GetPhi     ();
  GENPhase2 phaseCM(qcm, q2x1, q2x2, cmframe, cosx1, phix1, 0);
  ANL4DVector px1 = phaseCM.GetFourMomentum(0);
  ANL4DVector px2 = phaseCM.GetFourMomentum(1);
  Double_t betax  = phaseCM.GetBetaBar();
  if (betax <= 0.) return 0.;

  // X- ->tb b 
  GENBranch &x1branch = *cmbranch.GetBranchPtr(0);
  Double_t cost1 = x1branch.GetCosTheta();
  Double_t phit1 = x1branch.GetPhi     ();
  Double_t mt12  = x1branch.GetM12();
  Double_t m32   = x1branch.GetM22();
  GENPhase2 phaseHM(px1, mt12, m32, cmframe, cost1, phit1, 1);
  fP[0]           = phaseHM.GetFourMomentum(1);
  fM[0]           = TMath::Sqrt(m32);
  ANL4DVector pt1 = phaseHM.GetFourMomentum(0);
  Double_t betat1 = phaseHM.GetBetaBar();
  if (betat1 <= 0.) return 0.;

  // tb -> W- b
  GENBranch &t1branch = *x1branch.GetBranchPtr(0);
  Double_t cosw1   = t1branch.GetCosTheta();
  Double_t phiw1   = t1branch.GetPhi     ();
  Double_t mw12    = t1branch.GetM12();
  Double_t m72     = t1branch.GetM22();
  GENFrame x1frame = phaseHM.GetFrame();
  GENPhase2 phaseT1(pt1, mw12, m72, x1frame, cosw1, phiw1, 1);
  fP[1]           = phaseT1.GetFourMomentum(1);
  fM[1]           = TMath::Sqrt(m72);
  ANL4DVector pw1 = phaseT1.GetFourMomentum(0);
  Double_t betaw1 = phaseT1.GetBetaBar();
  if (betaw1 <= 0.) return 0.;

  // W- -> fub + fd
  GENBranch &w1branch = *t1branch.GetBranchPtr(0);
  Double_t cosf1   = w1branch.GetCosTheta();
  Double_t phif1   = w1branch.GetPhi     ();
  Double_t m112    = w1branch.GetM12();
  Double_t m122    = w1branch.GetM22();
  GENFrame t1frame = phaseT1.GetFrame();
  GENPhase2 phaseWM(pw1, m112, m122, t1frame, cosf1, phif1, 1);
  fP[2] = phaseWM.GetFourMomentum(0);
  fP[3] = phaseWM.GetFourMomentum(1);
  fM[2] = TMath::Sqrt(m112);
  fM[3] = TMath::Sqrt(m122);
  Double_t betaf1 = phaseWM.GetBetaBar();
  if (betaf1 <= 0.) return 0.;

  // X- ->tb b 
  GENBranch &x2branch = *cmbranch.GetBranchPtr(1);
  Double_t cost2 = x2branch.GetCosTheta();
  Double_t phit2 = x2branch.GetPhi     ();
  Double_t mt22  = x2branch.GetM12();
  Double_t m52   = x2branch.GetM22();
  GENPhase2 phaseHP(px2, mt22, m52, cmframe, cost2, phit2, 1);
  fP[4]           = phaseHP.GetFourMomentum(1);
  fM[4]           = TMath::Sqrt(m52);
  ANL4DVector pt2 = phaseHP.GetFourMomentum(0);
  Double_t betat2 = phaseHP.GetBetaBar();
  if (betat2 <= 0.) return 0.;

  // t -> W+ bb
  GENBranch &t2branch = *x2branch.GetBranchPtr(0);
  Double_t cosw2   = t2branch.GetCosTheta();
  Double_t phiw2   = t2branch.GetPhi     ();
  Double_t mw22    = t2branch.GetM12();
  Double_t m92     = t2branch.GetM22();
  GENFrame x2frame = phaseHP.GetFrame();
  GENPhase2 phaseT2(pt2, mw22, m92, x2frame, cosw2, phiw2, 1);
  fP[5]           = phaseT2.GetFourMomentum(1);
  fM[5]           = TMath::Sqrt(m92);
  ANL4DVector pw2 = phaseT2.GetFourMomentum(0);
  Double_t betaw2 = phaseT2.GetBetaBar();
  if (betaw2 <= 0.) return 0.;

  // W+ -> fu + fdb
  GENBranch &w2branch = *t2branch.GetBranchPtr(0);
  Double_t cosf2   = w2branch.GetCosTheta();
  Double_t phif2   = w2branch.GetPhi     ();
  Double_t m132    = w2branch.GetM12();
  Double_t m142    = w2branch.GetM22();
  GENFrame t2frame = phaseT2.GetFrame();
  GENPhase2 phaseWP(pw2, m132, m142, t2frame, cosf2, phif2, 1);
  fP[6] = phaseWP.GetFourMomentum(0);
  fP[7] = phaseWP.GetFourMomentum(1);
  fM[6] = TMath::Sqrt(m132);
  fM[7] = TMath::Sqrt(m142);
  Double_t betaf2 = phaseWP.GetBetaBar();
  if (betaf2 <= 0.) return 0.;

  Double_t eb     = rs/2.;
  Double_t pb     = TMath::Sqrt((eb-kM_e)*(eb+kM_e));
  Double_t beta_e = pb/eb;
  fK[0].SetXYZT(0., 0., pb, eb);
  fK[1].SetXYZT(0., 0.,-pb, eb);
#ifdef __DEBUG__
  cerr << "---------------------" << endl;
  cerr << " fP[0] = (" 
       << fP[0].E () << ","
       << fP[0].Px() << ","
       << fP[0].Py() << ","
       << fP[0].Pz() << ")" << endl;
  cerr << " fP[1] = (" 
       << fP[1].E () << ","
       << fP[1].Px() << ","
       << fP[1].Py() << ","
       << fP[1].Pz() << ")" << endl;
  cerr << " fP[2] = (" 
       << fP[2].E () << ","
       << fP[2].Px() << ","
       << fP[2].Py() << ","
       << fP[2].Pz() << ")" << endl;
  cerr << " fP[3] = (" 
       << fP[3].E () << ","
       << fP[3].Px() << ","
       << fP[3].Py() << ","
       << fP[3].Pz() << ")" << endl;
  cerr << " fP[4] = (" 
       << fP[4].E () << ","
       << fP[4].Px() << ","
       << fP[4].Py() << ","
       << fP[4].Pz() << ")" << endl;
  cerr << " fP[5] = (" 
       << fP[5].E () << ","
       << fP[5].Px() << ","
       << fP[5].Py() << ","
       << fP[5].Pz() << ")" << endl;
  cerr << " fP[6] = (" 
       << fP[6].E () << ","
       << fP[6].Px() << ","
       << fP[6].Py() << ","
       << fP[6].Pz() << ")" << endl;
  cerr << " fP[7] = (" 
       << fP[7].E () << ","
       << fP[7].Px() << ","
       << fP[7].Py() << ","
       << fP[7].Pz() << ")" << endl;
  ANL4DVector qw1 = fP[2] + fP[3];
  ANL4DVector qt1 = fP[1] + qw1;
  ANL4DVector qx1 = fP[0] + qt1;
  cerr << " qw1 = (" 
       << qw1.E () << ","
       << qw1.Px() << ","
       << qw1.Py() << ","
       << qw1.Pz() << ")" << endl;
  cerr << " qt1 = (" 
       << qt1.E () << ","
       << qt1.Px() << ","
       << qt1.Py() << ","
       << qt1.Pz() << ")" << endl;
  cerr << " qx1 = (" 
       << qx1.E () << ","
       << qx1.Px() << ","
       << qx1.Py() << ","
       << qx1.Pz() << ")" << endl;
  ANL4DVector qw2 = fP[6] + fP[7];
  ANL4DVector qt2 = fP[5] + qw2;
  ANL4DVector qx2 = fP[4] + qt2;
  cerr << " qw2 = (" 
       << qw2.E () << ","
       << qw2.Px() << ","
       << qw2.Py() << ","
       << qw2.Pz() << ")" << endl;
  cerr << " qt2 = (" 
       << qt2.E () << ","
       << qt2.Px() << ","
       << qt2.Py() << ","
       << qt2.Pz() << ")" << endl;
  cerr << " qx2 = (" 
       << qx2.E () << ","
       << qx2.Px() << ","
       << qx2.Py() << ","
       << qx2.Pz() << ")" << endl;
  cerr << " ---- " << endl;
  cerr << " rs = " << rs << endl;
  cerr << " mw1 = " << qw1.GetMass() << endl;
  cerr << " mw2 = " << qw2.GetMass() << endl;
  cerr << " mt1 = " << qt1.GetMass() << endl;
  cerr << " mt2 = " << qt2.GetMass() << endl;
  cerr << " mx1 = " << qx1.GetMass() << endl;
  cerr << " mx2 = " << qx2.GetMass() << endl;

  ANL4DVector pcm = qx1 + qx2;
  cerr << " pcm = (" 
       << pcm.E () << ","
       << pcm.Px() << ","
       << qcm.Py() << ","
       << qcm.Pz() << ")" << endl;
#endif

  // --------------------------------------------
  //  Calcuate differential cross section
  // --------------------------------------------
  Double_t s      = rs * rs;

  // -------------------
  //  Amplitude squared
  // -------------------
  Double_t amp2 = AmpSquared(cmbranch);

  // -------------------
  //  Put them together
  // -------------------
#ifndef __NODECAY__
  static const Int_t    kNbr  = 7;
  static const Double_t kFact = k2Pi/(TMath::Power(k4Pi,3*kNbr));
#else
  static const Double_t kFact = k2Pi/(TMath::Power(k4Pi,3+6));
  betat1 = 1.;
  betat2 = 1.;
  betaw1 = 1.;
  betaw2 = 1.;
  betaf1 = 1.;
  betaf2 = 1.;
#endif

  Double_t identp = 1.;                              // identical particle factor
  Double_t dPhase = kFact * betax * betat1 * betat2  // phase space factor
	                          * betaw1 * betaw2 
	                          * betaf1 * betaf2;
  Double_t flux   = 1./(2.* s * beta_e);             // beam flux factor
  Double_t spin   = 1./2.;                           // spin average for e+

  Double_t sigma  = identp * flux * spin * amp2 * dPhase; // in [1/GeV^2]
           sigma *= kGeV2fb;                              // now in [fb]

  return sigma;
}

//_____________________________________________________________________________
// --------------------------
//  AmpSquared()
// --------------------------
Double_t HCHCBases::AmpSquared(GENBranch &cmbranch)
{
#ifndef __NODECAY__    //                          _
  Double_t  cfh   = 3; // color factor for H+ -> t b
  Double_t  color = f1Ptr->GetColor() * f4Ptr->GetColor() * cfh * cfh;
  Int_t     ig1   = f1Ptr->GetGenNo() - 1;
  Int_t     ig2   = f2Ptr->GetGenNo() - 1;
  Int_t     ig4   = f4Ptr->GetGenNo() - 1;
  Int_t     ig5   = f5Ptr->GetGenNo() - 1;
  Double_t  vkm12 = f1Ptr->GetColor() > 1. ? kVkm[ig1][ig2] : 1.;
  Double_t  vkm45 = f4Ptr->GetColor() > 1. ? kVkm[ig4][ig5] : 1.;
  Double_t  mix   = TMath::Power(vkm12*vkm45,2);
#else
  Double_t  color = 1.;
  Double_t  mix   = 1.;
#endif
  Complex_t amp   = FullAmplitude();
  Double_t  amp2  = TMath::Power(abs(amp),2) * color * mix;

  return amp2;
}

//_____________________________________________________________________________
// --------------------------
//  FullAmplitude()
// --------------------------
Complex_t HCHCBases::FullAmplitude()
{
   Double_t gamw   = fWMBosonPtr->GetWidth();

   Double_t glw    = -kGw*kSqh;
   Double_t grw    = 0.;

   Double_t mx     = fHCBosonPtr->GetMass();
   Double_t gamx   = fHCBosonPtr->GetWidth();

   Double_t mb     = kMass[1][1][2];
   Double_t mt     = fT1Ptr->GetMass();
   Double_t gamt   = fT1Ptr->GetWidth();

   Double_t tanb   = GetTanBeta();
   Double_t cotb   = 1/tanb;

   Complex_t gltbm = kGw*kSqh/(2*kM_w) * (mb*tanb);
   Complex_t grtbm = kGw*kSqh/(2*kM_w) * (mt*cotb);
   Complex_t gltbp = grtbm;
   Complex_t grtbp = gltbm;

   static const Bool_t kIsIncoming = kTRUE;
   static const Bool_t kIsOutgoing = kFALSE;

   HELFermion em(fK[0], kM_e, fHelInitial[0], +1, kIsIncoming);
   HELFermion ep(fK[1], kM_e, fHelInitial[1], -1, kIsOutgoing);

#ifndef __NODECAY__
   HELFermion bxm(fP[0], fM[0], fHelFinal [0], +1, kIsOutgoing);
   HELFermion bb (fP[1], fM[1], fHelFinal [1], -1, kIsIncoming);
   HELFermion fub(fP[2], fM[2], fHelFinal [2], -1, kIsIncoming);
   HELFermion fd (fP[3], fM[3], fHelFinal [3], +1, kIsOutgoing);

   HELVector  wm(fub, fd , glw  , grw  , kM_w, gamw);
   HELFermion tb(bb , wm , glw  , grw  , mt  , gamt);
   HELScalar  xm(tb , bxm, gltbm, grtbm, mx  , gamx);

   HELFermion bxp(fP[4], fM[4], fHelFinal [4], -1, kIsIncoming);
   HELFermion b  (fP[5], fM[5], fHelFinal [5], +1, kIsOutgoing);
   HELFermion fu (fP[6], fM[6], fHelFinal [6], +1, kIsOutgoing);
   HELFermion fdb(fP[7], fM[7], fHelFinal [7], -1, kIsIncoming);

   HELVector  wp(fdb, fu, glw  , grw  , kM_w, gamw);
   HELFermion t (b  , wp, glw  , grw  , mt  , gamt);
   HELScalar  xp(bxp, t , gltbp, grtbp, mx  , gamx);
#ifdef __DEBUG__
   cerr << " wm : "; 
   wm.GetFourMomentum().DebugPrint();
   cerr << " wp : "; 
   wp.GetFourMomentum().DebugPrint();

   cerr << " xm = " << static_cast<Complex_t>(xm) << " ";
   xm.GetFourMomentum().DebugPrint();
   cerr << " xp = " << static_cast<Complex_t>(xp) << " ";
   xp.GetFourMomentum().DebugPrint();
#endif
#else
   ANL4DVector pxm = fP[0] + fP[1] + fP[2] + fP[3];
   ANL4DVector pxp = fP[4] + fP[5] + fP[6] + fP[7];
   HELScalar  xm(pxm, +1);
   HELScalar  xp(pxp, +1);
#endif

   Complex_t amp = AmpEEtoHH(em, ep, xm, xp);

   return amp;
}

//_____________________________________________________________________________
// --------------------------
//  AmpEEtoHH()
// --------------------------
Complex_t HCHCBases::AmpEEtoHH(const HELFermion &em,
                               const HELFermion &ep,
                               const HELScalar  &xm,
                               const HELScalar  &xp)
{
   //-------------------
   // Coupling consts.
   //-------------------
   Double_t  qe    = -1.;
   Double_t  ge    = -qe*kGe;
   Double_t  glae  = ge;
   Double_t  grae  = ge;

   Double_t  t3e   = -1./2.;
   Double_t  glze  = -kGz*(t3e - qe*kSin2W);
   Double_t  grze  = -kGz*(    - qe*kSin2W);

   Double_t  gamz  = fZBosonPtr->GetWidth();

   Double_t  gzxx  = glze;
   Double_t  gaxx  = ge;

   //--------------------------------------------------
   // Calculate Amplitudes
   //--------------------------------------------------
#ifdef __PHASESPACE__
   //---------------------------
   // Just Phase Space
   //---------------------------
   Complex_t amp = 1;
#else
   //---------------------------
   // XX Production Amplitude
   //---------------------------
   //--
   // S-channel Z
   //--
   HELVector zs(em, ep, glze, grze, kM_z, gamz); // s-channel   Z: backward-going
   HELVertex ampzxx(zs, xm, xp, gzxx);

   //--
   // S-channel A
   //--
   HELVector as(em, ep, glae, grae,   0.,   0.); // s-channel gamma
   HELVertex ampaxx(as, xm, xp, gaxx);

   //--
   // Sum of the two
   //--
   Complex_t amp = ampzxx + ampaxx;
#endif /* end __PHASESPACE__ */

   return amp;
}

//_____________________________________________________________________________
// --------------------------
//  Userin
// --------------------------
void HCHCBases::Userin()
{
  TDirectory *last = gDirectory;
  // --------------------------------------------
  //  Open beamstrahlung data
  // --------------------------------------------
  string bsfiledir;
  string bsfilename;

  if (fBeamStr) {
    stringstream ins(gJSF->Env()->GetValue("HCHCBases.BeamstrahlungFilepath",
                                           "/proj/soft/jsf/pro/data/bsdata/"));
    ins >> bsfiledir;

    ins.clear();
    ins.str(gJSF->Env()->GetValue("HCHCBases.BeamstrahlungFilename","trc500"));
    ins >> bsfilename;
    
    TFile *fBeamFile = new TFile((bsfiledir+bsfilename+".root").data());
    if (!fBeamFile) {
      cerr << " Unable to open a file for beamstrahlung" << endl;
      abort();
    }
    
  // --------------------------------------------
  //  Initialize beam generator
  // --------------------------------------------
    fBM = (JSFBeamGenerationCain*)fBeamFile->Get(bsfilename.data());
    fBM->SetIBParameters(fBeamWidth);

    fBM->MakeBSMap();
    fBM->Print(); 

    cout << " Nominal Energy (beamstrahlung) = " << fBM->GetNominalEnergy() 
         << " beam width     (beamstrahlung) = " << fBM->GetIBWidth() 
         << endl;

    if (!(fBM->GetNominalEnergy() == (fEcmInit/2))) {
      cout << "Nominal energy from beamstrahung is " << fBM->GetNominalEnergy() 
           << " which is different from fEcm/2: "    << (fEcmInit/2)
           << endl;        
    } // check the energy homogeneity
  } // if beamstrahlung is on
  last->cd();

  // --------------------------------------------
  //  Initialize Z decay table
  // --------------------------------------------
  if (!fWMBosonPtr) fWMBosonPtr = new GENPDTWBoson();
  for (Int_t m=1; m<=fWMBosonPtr->GetEntries(); m++) {
     GENDecayMode *mp = fWMBosonPtr->GetMode(m); 
     if (mp && (m<fWmModesLo || m>fWmModesHi)) {
        mp->Lock();
     }
  }
  fWMBosonPtr->DebugPrint();
  if (!fWPBosonPtr) fWPBosonPtr = new GENPDTWBoson();
  for (Int_t m=1; m<=fWPBosonPtr->GetEntries(); m++) {
     GENDecayMode *mp = fWPBosonPtr->GetMode(m); 
     if (mp && (m<fWpModesLo || m>fWpModesHi)) {
        mp->Lock();
     }
  }
  fWPBosonPtr->DebugPrint();
  if (!fZBosonPtr) fZBosonPtr = new GENPDTZBoson();
  fZBosonPtr->DebugPrint();
  if (!fPhotonPtr) fPhotonPtr = new GENPDTPhoton();
  //fPhotonPtr->DebugPrint();
  if (!fGluonPtr)  fGluonPtr  = new GENPDTGluon();
  //fGluonPtr->DebugPrint();

  Double_t mb = kMass[1][1][2];
  if (!fHCBosonPtr)   fHCBosonPtr  = new HCBoson(fMass,fTanBeta,fTopMass);
  fHCBosonPtr->DebugPrint();

  Double_t mt = fTopMass;
  if (!fT1Ptr)  fT1Ptr= new TopQuark(mt);
  fT1Ptr->DebugPrint();

  if (!fT2Ptr)  fT2Ptr= new TopQuark(mt);
  fT2Ptr->DebugPrint();

  Double_t mx = fMass;

  // --------------------------------------------
  //  Define some plots
  // --------------------------------------------
  Xh_init( 1,      0., fEcmInit*1.1,  50, "Ecm"    );
  Xh_init( 2, fXL[ 0], fXU[ 0],       50, "Costh"  );
  Xh_init( 3, fXL[ 1], fXU[ 1],       50, "Phi"    );
  Xh_init( 4,  mx-10.,  mx+10.,       50, "Mx-"    );
  Xh_init( 5, fXL[ 2], fXU[ 2],       50, "CosthT1");
  Xh_init( 6, fXL[ 3], fXU[ 3],       50, "PhiT1"  );

  Xh_init( 7,    120.,    220.,       50, "Mt1"    );
  Xh_init( 8, fXL[ 4], fXU[ 4],       50, "CosthW-");
  Xh_init( 9, fXL[ 5], fXU[ 5],       50, "PhiW-"  );

  Xh_init(10,     60.,    100.,       50, "MW-"    );
  Xh_init(11, fXL[ 6], fXU[ 6],       50, "CosthF1");
  Xh_init(12, fXL[ 7], fXU[ 7],       50, "PhiF1"  );

  Xh_init(13,  mx-10.,  mx+10.,       50, "Mx+"    );
  Xh_init(14, fXL[ 8], fXU[ 8],       50, "CosthT2");
  Xh_init(15, fXL[ 9], fXU[ 9],       50, "PhiT2"  );

  Xh_init(16,    120.,    220.,       50, "Mt2"    );
  Xh_init(17, fXL[10], fXU[10],       50, "CosthW+");
  Xh_init(18, fXL[11], fXU[11],       50, "PhiW+"  );

  Xh_init(19,     60.,    100.,       50, "MW+"    );
  Xh_init(20, fXL[12], fXU[12],       50, "CosthF2");
  Xh_init(21, fXL[13], fXU[13],       50, "PhiF2"  );
  Xh_init(22,      0.,      2.,        2, "Helin " );
  Xh_init(23,      0.,      4.,        4, "Helot " );
  Xh_init(24,      0.,     12.,       12, "W- mode");
  Xh_init(25,      0.,     12.,       12, "W+ mode");
#ifdef TEMP_H
  H1Init("h02", "cos(theta_X)",200,-1.,+1.0);
#endif
}

//_____________________________________________________________________________
// --------------------------
//  Userout
// --------------------------
void HCHCBases::Userout()
{
  cout << "End of HCHCBases----------------------------------- "  << endl
       << "Ecm                  = " << fEcmInit << " [GeV]   "    << endl
       << "Beamstrahlung        = " << (fBeamStr ? "on" : "off")  << endl
       << "Bremsstrahlung       = " << (fISR     ? "on" : "off")  << endl
       << "Total Cross section  = " << GetEstimate()  << " +/- "
                                    << GetError()     << " [fb]"  << endl
       << "Number of iterations = " << GetNoOfIterate()           << endl;
}

//_____________________________________________________________________________
// --------------------------
//  SelectHelicities
// --------------------------
void HCHCBases::SelectHelicities(Double_t &weight)
{
   static const Int_t kNi = 2;
   static const Int_t kIHelComb[kNi][2] = {{-1, +1},
                                           {+1, -1}};
   static const Int_t kNf = 4;
   static const Int_t kFHelComb[kNf][8] = {{-1, +1,+1,-1,  -1, -1,-1,+1},
                                           {-1, +1,+1,-1,  +1, -1,-1,+1},
                                           {+1, +1,+1,-1,  -1, -1,-1,+1},
                                           {+1, +1,+1,-1,  +1, -1,-1,+1}};
   Double_t helm = (1. - fPole)/2.;
   if (fHelCombInitial < helm) {
      fJCombI = 0;
   } else {
      fJCombI = 1;
   }
   fHelInitial[0] = kIHelComb[fJCombI][0];
   fHelInitial[1] = kIHelComb[fJCombI][1];
   fJCombF = (Int_t)(fHelCombFinal*kNf);
   fJCombF = TMath::Min(fJCombF, kNf-1);
   fHelFinal  [0] = kFHelComb[fJCombF][0];
   fHelFinal  [1] = kFHelComb[fJCombF][1];
   fHelFinal  [2] = kFHelComb[fJCombF][2];
   fHelFinal  [3] = kFHelComb[fJCombF][3];
   fHelFinal  [4] = kFHelComb[fJCombF][4];
   fHelFinal  [5] = kFHelComb[fJCombF][5];
   fHelFinal  [6] = kFHelComb[fJCombF][6];
   fHelFinal  [7] = kFHelComb[fJCombF][7];
   weight = kNf;
#ifdef __NODECAY__
   weight = 1.;
#endif
}

//-----------------------------------------------------------------------------
// ==============================
//  class HCBoson
// ==============================
//_____________________________________________________________________________
// --------------------------
//  C-tor
// --------------------------
HCBoson::HCBoson(Double_t m,
                 Double_t tanb,
                 Double_t mt)
        : fTanBeta   (tanb),
          fTopMass   (mt)
{
   fName    = TString("H+");
   fPID     =   37;
   fCharge  =  1.0;
   fSpin    =  0.0;
   fMass    =    m;
   fGen     =    0;
   fIsoSpin =  0.5;
   fColor   =  1.0;

   Initialize();
}

//_____________________________________________________________________________
// --------------------------
//  Initialize
// --------------------------
void HCBoson::Initialize()
{
   //--
   // Couplings
   //--
   Double_t tanb   = fTanBeta;
   Double_t cotb   = 1/tanb;
   Double_t mt     = fTopMass;
   Double_t mb     = kMass[1][1][2];

   Complex_t gl = kGw*kSqh/(2*kM_w) * (mt*cotb);
   Complex_t gr = kGw*kSqh/(2*kM_w) * (mb*tanb);
   //--          _
   // H+ --> t + b
   //--
   int ic = 1;
   int it = 1;
   int ig = 2;
   const Char_t   *name = kName[ic][it][ig];
         Int_t     pid  = kPID [ic][it][ig];  
         Double_t  t3   = (1 - 2*it)/2.;
         Double_t  qf   = kChrg[ic][it][ig];
         Double_t  spin = 0.5;
         Double_t  cb   = 2*ic + 1;
         Double_t  mass = kMass[ic][it][ig];
   GENDecayMode *dmp;
   GENPDTEntry  *d1p, *d2p;
   d1p = new TopQuark(fTopMass);
   d2p = new GENPDTEntry(name,-pid,-qf, spin, mass, ig+1, -t3, cb);
   Double_t cf    = 3.;
   Double_t m1    = d1p->GetMass();
   Double_t m2    = d2p->GetMass();
   Double_t ident = 1.;
   Double_t gam = GamToFF(m1, m2, gl, gr, cf)/ident;
   if (gam > 0.) {
      dmp = new GENDecayMode(gam);
      dmp->Add(d1p);
      dmp->Add(d2p);
      Add(dmp);
   }
}

//_____________________________________________________________________________
// --------------------------
//  GamToFF
// --------------------------
Double_t HCBoson::GamToFF(Double_t  m1, // 1st daughter mass
                          Double_t  m2, // 2nd daughter mass
                          Complex_t gl, // left  coupling
                          Complex_t gr, // right coupling
                          Double_t  cf) // color factor
{
   Double_t x1   = TMath::Power(m1/fMass,2);
   Double_t x2   = TMath::Power(m2/fMass,2);
   Double_t beta = 1. - 2.*(x1+x2) + TMath::Power((x1-x2),2);

   if (beta <= 0.) return 0.;
   beta = TMath::Sqrt(beta);

   ANL4DVector px(fMass,0.,0.,0.);
   double p = fMass*beta/2;
   ANL4DVector pt(sqrt(p*p+m1*m1),0.,0., p);
   ANL4DVector pb(sqrt(p*p+m2*m2),0.,0.,-p);
   HELScalar  x(px, -1);

   static const Bool_t kIsIncoming = kTRUE;
   static const Bool_t kIsOutgoing = kFALSE;

   HELFermion tl (pt, m1, -1, +1, kIsOutgoing);
   HELFermion bbl(pb, m2, -1, -1, kIsIncoming);
   HELFermion tr (pt, m1, +1, +1, kIsOutgoing);
   HELFermion bbr(pb, m2, +1, -1, kIsIncoming);

   HELVertex ampl(bbl, tl, x, gl, gr);
   HELVertex ampr(bbr, tr, x, gl, gr);

   double tta = pow(abs(ampl),2) + pow(abs(ampr),2);

   Double_t fac = 1./(16.*kPi)/fMass;
   Double_t gam = fac*tta*beta*cf;

   return gam;
}

//-----------------------------------------------------------------------------
// ==============================
//  class TopQuark
// ==============================
//_____________________________________________________________________________
// --------------------------
//  C-tor
// --------------------------
TopQuark::TopQuark(Double_t m)
{
   fName    = TString("top");
   fPID     =    6;
   fCharge  = 2./3;
   fSpin    =  0.5;
   fMass    =    m;
   fGen     =    3;
   fIsoSpin =  0.5;
   fColor   =  3.0;

   Initialize();
}

//_____________________________________________________________________________
// --------------------------
//  Initialize
// --------------------------
void TopQuark::Initialize()
{
   //--
   // Couplings
   //--
   Double_t mt     = fMass;
   Double_t mb     = kMass[1][1][2];

   Double_t glw    = -kGw*kSqh;
   Double_t grw    = 0.;
   //--      
   // t --> b + W+
   //--
   int ic = 1;
   int it = 1;
   int ig = 2;
   const Char_t   *name = kName[ic][it][ig];
         Int_t     pid  = kPID [ic][it][ig];  
         Double_t  t3   = (1 - 2*it)/2.;
         Double_t  qf   = kChrg[ic][it][ig];
         Double_t  spin = 0.5;
         Double_t  cb   = 2*ic + 1;
         Double_t  mass = kMass[ic][it][ig];
   GENDecayMode *dmp;
   GENPDTEntry  *d1p, *d2p;
   d1p = new GENPDTEntry(name,pid,qf, spin, mass, ig+1, t3, cb);
   d2p = new GENPDTWBoson();
   Double_t vkm   = kVkm[2][2];
   Double_t m1    = d1p->GetMass();
   Double_t m2    = d2p->GetMass();
   Double_t ident = 1.;
   Double_t gam = GamToFV(m1, m2, glw, grw, vkm)/ident;
   if (gam > 0.) {
      dmp = new GENDecayMode(gam);
      dmp->Add(d1p);
      dmp->Add(d2p);
      Add(dmp);
   }
}

//_____________________________________________________________________________
// --------------------------
//  GamToFF
// --------------------------
Double_t TopQuark::GamToFV(Double_t  m1,  // daughter fermion mass
                           Double_t  m2,  // Vector mass
			   Double_t  gl,  // left  coupling
			   Double_t  gr,  // right coupling
			   Double_t  vkm) // Vkm elment
{
   Double_t x1   = TMath::Power(m1/fMass,2);
   Double_t x2   = TMath::Power(m2/fMass,2);
   Double_t beta = 1. - 2.*(x1+x2) + TMath::Power((x1-x2),2);

   if (beta <= 0.) return 0.;
   beta = TMath::Sqrt(beta);

   ANL4DVector pt(fMass,0.,0.,0.);
   double p = fMass*beta/2;
   ANL4DVector pb(sqrt(p*p+m1*m1),0.,0.,p);
   ANL4DVector pw(sqrt(p*p+m2*m2),0.,0.,-p);

   static const Bool_t kIsIncoming = kTRUE;
   static const Bool_t kIsOutgoing = kFALSE;

   HELFermion tl(pt, fMass, -1, +1, kIsIncoming);
   HELFermion tr(pt, fMass, +1, +1, kIsIncoming);
   HELFermion b (pb, m1   , -1, +1, kIsOutgoing);
   HELVector  wm(pw, m2, -1, +1);
   HELVector  w0(pw, m2,  0, +1);
   HELVector  wp(pw, m2, +1, +1);

   HELVertex amplm(tl, b, wm, gl, gr);
   HELVertex ampl0(tl, b, w0, gl, gr);
   HELVertex amplp(tl, b, wp, gl, gr);

   HELVertex amprm(tr, b, wm, gl, gr);
   HELVertex ampr0(tr, b, w0, gl, gr);
   HELVertex amprp(tr, b, wp, gl, gr);
   
   double tta = pow(abs(amplm),2) + pow(abs(ampl0),2)+ pow(abs(amplp),2)
              + pow(abs(amprm),2) + pow(abs(ampr0),2)+ pow(abs(amprp),2);

   Double_t spf  = 2;
   Double_t fac  = 1./(16.*kPi)/fMass/spf;
   Double_t gam  = fac*tta*beta*vkm*vkm;

   return gam;
}
