//*************************************************************************
//* ---------------------
//* S4ITHit Class
//* ---------------------
//* (Description)
//*	Class for a single IT hit.
//* (Requires)
//*	TObject
//* (Provides)
//*	S4ITHit		: TObject containing single hit data
//* (Update Record)
//*	2004/08/26  A.Yamaguchi	Original version.
//*************************************************************************

#include <iostream>
#include "S4ITHit.h"
#include "TRandom.h"
using namespace std;


ClassImp(S4ITHit)

Int_t    S4ITHit::fgSortBy = 0; 
Double_t S4ITHit::fgDRPhi  = 1.e-3;  // [cm]
Double_t S4ITHit::fgDZ     = 1.e-3;  // [cm]
Double_t S4ITHit::fgDX     = 1.e-3;  // [cm]
Double_t S4ITHit::fgDY     = 1.e-3;  // [cm]

//=====================================================================
//* constructor -------------------------------------------------------

S4ITHit::S4ITHit(S4ITExactHit *ehitp)
       : U4ITHit(ehitp->GetLayerID(),
                 ehitp->GetSide(),
                !ehitp->GetSide() ? ehitp->GetLayerR() * ehitp->GetPhi()
                                  : ehitp->GetPosition().X(),
                !ehitp->GetSide() ? ehitp->GetPosition().Z()
                                  : ehitp->GetPosition().Y()),
         fExactHitPtr(ehitp)
{ 
   ehitp->SetSmearedHitPtr(this);
}

//=====================================================================
//* destructor --------------------------------------------------------

S4ITHit::~S4ITHit() 
{
}
  
//=====================================================================
//* SetSortBy ------------------------------------------------------=--

void S4ITHit::SetSortBy(Char_t *opt)
{
   switch (opt[2]) {
      case 'a':    // sort by "TrackID"
         fgSortBy = 0;
         break;
      case 'm':    // sort by "Time"
      default:
         fgSortBy = 1;
         break;
   }
}

//=====================================================================
//* SetSortBy ------------------------------------------------------=--

Int_t S4ITHit::Compare(const TObject *obj) const
{
   switch (fgSortBy) {
      case 0: {
         Int_t trackid  = fExactHitPtr->GetTrackID();
         Int_t trackido = dynamic_cast<const S4ITHit *>(obj)
                                           ->fExactHitPtr->GetTrackID();
         return (trackid < trackido ? -1 : (trackid > trackido ? 1 : 0));
      }
      case 1: 
      default: {
         Double_t tof  = fExactHitPtr->GetTOF();
         Double_t tofo = dynamic_cast<const S4ITHit *>(obj)
                                          ->fExactHitPtr->GetTOF();
         return (tof < tofo ? -1 : (tof > tofo ? 1 : 0));
      }
   }
}

//=====================================================================
//* Update ------------------------------------------------------------

void S4ITHit::Update()
{
   U4ITHit::Update();
}

//=====================================================================
//* DebugPrint --------------------------------------------------------

void S4ITHit::DebugPrint(const Char_t *)
{
   if (!IsDone()) Update();

   cerr << right << fixed
        << "--------------------------------------------------" << endl
        << " Track ID = " << fExactHitPtr->GetTrackID()         << endl
        << " Etot     = " << fExactHitPtr->GetTotalEnergy()     << endl
        << " LayerID  = " << GetLayerID()                       << endl
        << " Side     = " << GetSide()                          << endl
        << " X1       = " << GetX1() << " +/- " << GetDX1()     << endl
        << " X2       = " << GetX2() << " +/- " << GetDX2()     << endl;
}

//=====================================================================
//* CalculatePosition() -----------------------------------------------

void S4ITHit::CalculatePosition()
{
   if (!GetSide()) {
      SetDX1(fgDRPhi);
      SetX1(fExactHitPtr->GetLayerR() * fExactHitPtr->GetPhi()
                                      + gRandom->Gaus(0., fgDRPhi));
      SetDX2(fgDZ);
      SetX2(fExactHitPtr->GetPosition().Z() + gRandom->Gaus(0., fgDZ)); 
   } else {
      SetDX1(fgDX);
      SetX1(fExactHitPtr->GetPosition().X() + gRandom->Gaus(0., fgDX));
      SetDX2(fgDY);
      SetX2(fExactHitPtr->GetPosition().Y() + gRandom->Gaus(0., fgDY)); 
   }
}

//=====================================================================
//* Print () -----------------------------------------------

void S4ITHit::Print()
{
  cerr << "S4ITHit..Parameters for IT Hit point making " << endl;
  cerr << "     Sigma_{DRPhi} = " << fgDRPhi << " cm " << endl;
  cerr << "     Sigma_{DZ}    = " << fgDZ << " cm " << endl;
  cerr << "     Sigma_{DX}    = " << fgDX << " cm " << endl;
  cerr << "     Sigma_{DY}    = " << fgDY << " cm " << endl;

}
