//*************************************************************************
//* -----------
//* QQAnalysis
//* -----------
//* (Description)
//* 	Clasess to analyze Jet data.
//* (Update Record)
//*	2004/11/03  K.Fujii	Original version.
//*	2004/12/08  S.Yamamoto	WW version.
//*************************************************************************

#include "JSFSteer.h"
#include "QQAnalysis.h"
#include <iostream>
#include <sstream>

#include "S4GENMCTruthBuf.h"
#include "S4PFObject.h"
#include "S4PFOMakerBuf.h"
#include "S4JetMakerBuf.h"
#include "S4JetSet.h"
#include "S4Jet.h"

#include "ANLPairCombiner.h"

#include "TNtupleD.h"

ClassImp(QQAnalysis)

//___________________________________________________________________________
QQAnalysis::QQAnalysis(const Char_t *name, const Char_t *title)
             : JSFModule(name, title)
{
   //std::cerr << " constructor for QQAnalysis is called..." << std::endl;
}

//___________________________________________________________________________
QQAnalysis::~QQAnalysis()
{
}

//___________________________________________________________________________
Bool_t QQAnalysis::Initialize()
{
   TDirectory *last = gDirectory;
   gFile->cd("/");

   last->cd();

   return kTRUE;
}


//___________________________________________________________________________
Bool_t QQAnalysis::Process(Int_t ev)
{
   Monitor2Jets(ev);

#if 1
   std::cerr << "QQAnalysis::Process evt " << ev << " is finished" 
             << std::endl;
#endif
   return kTRUE;   
}   

static const Double_t kMassW   =  80.4; // W mass by s.y
static const Double_t kSigmaMw =   4.0; // W mass resolution
static const Double_t kMassZ   = 91.187; // Z mass
static const Double_t kSigmaMz =   4.0; // Z mass resolution
static const Double_t kMassH   = 120.0; // H mass
static const Double_t kSigmaMh =   4.0; // H mass resolution

static const Double_t xEvis    = 220.0; // Evis cut [GeV}
static const Double_t xPt      = 999.0; // Pt   cut [GeV}
static const Double_t xPl      = 999.0; // Pl   cut [GeV]
static const Double_t xM2j     =  40.0; // mass cut [GeV]
#if 0
static const Double_t xCosj    =   0.8; // cos(theta_J) cut
static const Int_t    xNpsj    =    10; // cut on #particles/jet
#else
static const Double_t xCosj    =  0.95; // cos(theta_J) cut
static const Int_t    xNpsj    =     0; // cut on #particles/jet
#endif
static       Int_t    ngood2   =     0;

//___________________________________________________________________________
void QQAnalysis::Monitor2Jets(Int_t ev)
{
   using namespace std;

   TDirectory *last = gDirectory;
   gFile->cd("/");

   // Prepare cut table.
   static TH1D *hStat2 = 0;
   if (!hStat2) hStat2 = new TH1D("hStat2", "Cut Table for 2 Jets", 20, 0, 20);
   Int_t cutid = 0;
   hStat2->Fill(cutid++,1.);

   // Get Jet buffer
   S4JetMakerBuf *jmbp = static_cast<S4JetMakerBuf *>(
                            gJSF->FindModule("S4JetMaker")->EventBuf()
                         );

   // Get jets with default ycut.
   Int_t njets = 0;
   S4JetSet      *jsp  = jmbp->GetJetSetPtr(njets);
   if (!jsp) return;
   hStat2->Fill(cutid++,1.);

   ANL4DVector qsum;
   TIter next(jsp);
   S4Jet *jp = 0;
   
   while ((jp = static_cast<S4Jet *>(next()))) qsum += *jp;

   Double_t evis = qsum(0);      // Evis
   Double_t pt   = qsum.GetPt(); // P_t
   Double_t pl   = qsum(3);      // P_l

   // Evis, Pt, and Pl cuts.
   static TH1D *hEvis2 = 0;
   if (!hEvis2) hEvis2 = new TH1D("hEvis2", "Evis for 2 Jets", 500, 0, 500);
   hEvis2->Fill(evis,1.);
   if (evis > xEvis) return;
   hStat2->Fill(cutid++,1.);

   static TH2D *hPtPl2 = 0;
   if (!hPtPl2) hPtPl2 = new TH2D("hPtPl2", "(Pt,Pl) for 2 Jets", 
                                          100,   0., 100., 100, -50.,  50.);
   hPtPl2->Fill(pt,pl,1.);
   if (pt > xPt || TMath::Abs(pl) > xPl) return;
   hStat2->Fill(cutid++,1.);

   // Select 2 jets.
   njets = 2;
   jsp  = jmbp->GetJetSetPtr(njets);
   if (!jsp) return;
   hStat2->Fill(cutid++,1.);

   static TH1D *hEjet2 = 0;
   if (!hEjet2) hEjet2 = new TH1D("hEjet2", "Ejet for 2 Jets", 200, 0, 200);

   static TH1D *hCosj2 = 0;
   if (!hCosj2) hCosj2 = new TH1D("hCosj2", "Cosj for 2 Jets", 200, -1., 1.);

   static TH1D *hNpsj2 = 0;
   if (!hNpsj2) hNpsj2 = new TH1D("hNpsj2", "Np/Jet for 2 Jets", 100, 0, 100);

   Int_t njok = 0;
   TIter nextjet(jsp);
   while ((jp = static_cast<S4Jet *>(nextjet()))) {
      hEjet2->Fill(jp->E(),1.);
      hCosj2->Fill(jp->CosTheta(),1.);
      hNpsj2->Fill(jp->GetNparticles(),1.);
      if (TMath::Abs(jp->CosTheta()) < xCosj 
          && jp->GetNparticles() > xNpsj) njok++;
   }
   if (njok < 2) return;
   hStat2->Fill(cutid++,1.);

   static TH2D *hMzMm2 = 0;
   if (!hMzMm2) hMzMm2 = new TH2D("hMzMm2", "(Mz,Mm) for 2 Jets",
                                            4000, 0., 200., 100,  -100., 400.);

   static TH2D *hCosMz = 0;
   if (!hCosMz) hCosMz = new TH2D("hCosMz", "(Cosj,Mz) for 2 Jets",
                                            200,  -1., 1., 4000,  0., 200.);

   // Mass cuts to select Z(,missing mass) candidates.
   
   S4JetSet &jets = *jsp;
   ANLPairCombiner zcandidates(jets,jets);
   ANLPair &z = *static_cast<ANLPair *>(zcandidates());

   S4GENMCTruthBuf *genbp = static_cast<S4GENMCTruthBuf *>(
                               gJSF->FindModule("S4GENMCTruth")->EventBuf()
                            );
   Double_t Ecm = genbp->GetPVcm().E();

   ANL4DVector qcm(Ecm);
   Double_t zmass = z().GetMass();
   ANL4DVector mm = qcm - z;
   Double_t mmass = mm.GetMass();
#if 1
   cerr << " M_Z = " << zmass << " M_m = " << mmass
        << " P_m = (" << mm.E() << ", " << mm.Px() << ", " << mm.Py() << ", " << mm.Pz() << ")" << endl;
#endif
   ANLJet *jet1p = static_cast<ANLJet *>(jets[0]);
   ANLJet *jet2p = static_cast<ANLJet *>(jets[1]);
   hMzMm2->Fill(zmass, mmass, 1.);
   hCosMz->Fill(jet1p->CosTheta(), zmass, 1.);

   if (TMath::Abs(zmass - kMassZ) > xM2j || TMath::Abs(mmass) > xM2j) return;
   hStat2->Fill(cutid++,1.);

   // Found a solution.
   // Do further analysis here.

   static TNtupleD *tupjetp = 0;
   if (!tupjetp) tupjetp = new TNtupleD("hjet","QQAnalysis","evis:pt:pl:cosj1:cosj2:ej1:ej2:npj1:npj2:mmass:zmass");
   tupjetp->Fill(evis,
		 pt,
		 pl,
		 jet1p->CosTheta(),
		 jet2p->CosTheta(),
		 jet1p->E(),
		 jet2p->E(),
		 jet1p->GetNparticles(),
		 jet2p->GetNparticles(),
		 mmass,
		 zmass);
   
   // Count up good events.
   ngood2++;
   // That's it, folks!
   last->cd();
}

//___________________________________________________________________________
void QQAnalysis::Draw(Option_t/* *opt */)
{
   gFile->cd("/");
   gDirectory->cd();
}
