#ifndef __S4PFOBJECT__
#define __S4PFOBJECT__

//*************************************************************************
//* ---------------------
//* S4PFObject Class
//* ---------------------
//* (Description)
//*	represent Particle Flow Object. 
//* (Requires)
//*	S4CALCluster, S4HYBTrack
//* (Provides)
//*	S4PFObject		: made object class
//* (Update Record)
//*	2004/09/04  S.Yamamoto	Original version.
//*	2005/07/17  S.Yamamoto	neutrino version.
//*************************************************************************

#include "ANL4DVector.h"
#include "S4HYBTrack.h"
#include "S4TPCPostHit.h"
#include "S4CALCluster.h"
#include "S4CALTypeDef.h"
#include "TAttDrawable.h"
#include "S4PrimaryParticle.h"
#include "S4BreakPoint.h"

class S4PFObject : public ANL4DVector, public TAttDrawable {
friend class S4PFOMakerBuf;
friend class S4V0Finder;

public:
 S4PFObject(S4CALCluster *cluster = 0);
 S4PFObject(S4HYBTrack   *track);
 S4PFObject(S4BreakPoint *bpp);
 S4PFObject(TObjArray    *childrenp);

 virtual ~S4PFObject();

 // Data Getters
 virtual Double_t      GetCharge();  
 virtual Double_t      GetProbability(EPID_t pid);
 virtual Int_t         GetPID();
 virtual Bool_t        IsPID(EPID_t pid, Double_t prob = 0.8);
 virtual Int_t         GetPDGCode(); 
 virtual TVector3      GetPosition(Int_t ip = 0); // (0,1) = (start,end)
 virtual Double_t      GetTime    (Int_t ip = 0); // (0,1) = (start,end) 
 virtual Double_t      GetConeEnergy(const Double_t cth, const TObjArray *tracks); 
 virtual S4CALCluster *GetClusterPtr     () { return fClusterPtr;      }
 virtual S4HYBTrack   *GetHYBTrackPtr    () { return fTrackPtr;        }
 virtual TObjArray    *GetDaughterPFOsPtr() { return fDaughterPFOsPtr; }
 virtual Double_t      GetQuality        () { return fQuality;         }

 virtual S4VExactHit  *GetExactHitPtr();
 virtual S4BreakPoint *GetPrimaryParticleBreakPoint();
 virtual S4BreakPoint *GetPrimaryBreakPointPtr();
 virtual S4BreakPoint *GetBreakPointPtr() { return fBreakPointPtr; }

 virtual void DebugPrint(const Char_t *opt = "");

 using TAttDrawable::Draw;
 virtual void Draw(Int_t color, const Char_t *opt = "");

 static  Bool_t   IsExact        (const Char_t *type);
 static  void     SetIsExact     (const Char_t *type, Bool_t set = kTRUE);
 static  Bool_t   GetUsePrimaryBP();
 static  void     SetUsePrimaryBP(Bool_t set = kTRUE);

private:
 virtual void     Update();
 virtual void     DoPID();
 virtual Double_t GetMass();
 virtual void     UpdateV0                   (TVector3 &p, Double_t &e, Double_t &qual);
 virtual void     UpdateWithPrimaryBreakPoint(TVector3 &p, Double_t &e);
 virtual void     UpdateWithBreakPoint       (TVector3 &p, Double_t &e);
 virtual void     SetPosition(const TVector3 &xv, Int_t ip = 0); // (0,1) = (start,end)

private:
 // **** TEMPORARY TRAETMENT ****
 static const Int_t    fkgNpid = 10;          //! Number of Particle Types
 static const Double_t fkgMasses[fkgNpid];    //! Particle Masses 
 // **** End of TEMPORARY TREATMENT ****

 S4CALCluster *fClusterPtr;           // pointer to cluster
 S4HYBTrack   *fTrackPtr;             // pointer to track
 S4TPCPostHit *fPostHitPtr;           // pointer to posthit
 S4BreakPoint *fBreakPointPtr;        // pointer to breakpoint
 TObjArray    *fDaughterPFOsPtr;      // pointer to an array of daughter PFOs
 Double_t      fProbs[fkgNpid];       // probabilities of PID 
 Int_t         fPID;                  // most probable pid
 Int_t         fPdgcode;              // exact pid
 TVector3      fXstart;               // strat point(cm)
 Double_t      fTstart;               // start time(nsec)
 Double_t      fMass;                 // mass of PFO(GeV)
 Double_t      fQuality;              // quality (meaningful only for V0s)
 Bool_t        fDone;	              // (1,0) = (pfo done, not yet)

 static Bool_t fgIsTKExact;           //! use mctruth for track if true
 static Bool_t fgIsEMExact;           //! use mctruth for EMCAL if true
 static Bool_t fgIsHDExact;           //! use mctruth for HDCAL if true
 static Bool_t fgUsePrimaryBP;        //! use primary BP info as mctruth

 ClassDef(S4PFObject, 1) // PFO data information
};

#endif
