//*************************************************************************
//* ---------------------
//* U4CDCHitWire Class
//* ---------------------
//* (Description)
//*	Class for a single CDC hit.
//* (Requires)
//*	TObject
//* (Provides)
//*	U4CDCHitWire		: Array containing "U4CDCHit"s
//* (Update Record)
//*	1999/12/04  K.Fujii	Original version.
//*************************************************************************

#include <iostream>
#include <iomanip>
#include <fstream>
#include "U4CDCHitWire.h"


ClassImp(U4CDCHitWire)
U4CDCHitWire::U4CDCHitWire(U4CDCTimeToXRelation *calib) : TObjArray(200),
		fLayerNumber(0), fCellNumber(0), fWireNumber(0), 
		fXwire(0.), fYwire(0.), fCellPtr(0),
		fCalibPtr(calib) 
{
}

U4CDCHitWire::~U4CDCHitWire()
{
	Delete();
}

Bool_t U4CDCHitWire::IsSortable() const { return kTRUE; }

Int_t U4CDCHitWire::Compare(const TObject *obj) const
{
	Int_t wire  = fWireNumber;
	Int_t wireo = ((U4CDCHitWire *)obj)->fWireNumber;
   	if (wire < wireo) return -1;
   	else if (wire > wireo) return 1;
   	else return 0;
}

Int_t U4CDCHitWire::Append(U4CDCHit *ht)
{
	Int_t 	 layer, cell, wire;
	Double_t xwire, ywire;
	ChannelToWire(ht->GetClusterPtr(),layer,cell,wire,xwire,ywire);
	
	// Return value = -1:	1st hit.
	//		=  0:	successfully appended.
	//		=  1:	new wire.
	//		=  2:	new cell.
	//		=  3:	new layer.
	
	if (GetEntries() == 0) {
		fLayerNumber  = layer;
		fCellNumber   = cell;
		fWireNumber   = wire;
#if 0
		std::cerr << " layer = " << fLayerNumber
			  << " cell = "  << fCellNumber
			  << " wire = "  << fWireNumber << std::endl;
#endif
		fXwire	  = xwire;
		fYwire	  = ywire;
		ht->fWirePtr = this;
                Add(ht);
                return -1;
	} else {
		if (layer == fLayerNumber) {
			if (cell == fCellNumber) {
                    		if (wire == fWireNumber) {
                    			ht->fWirePtr = this;
                        		Add(ht);
                        		return 0;
                        	} else {
                        		return 1;
                        	}
                        } else {
                        	return 2;
                        }
                } else {
                        return 3;
                }
        }
}

void U4CDCHitWire::DebugPrint(const Char_t *opt)
{
	std::cerr.setf(std::ios::right);
	std::cerr.setf(std::ios::fixed);
	std::cerr	<< "wire = (" << std::setw(1) << fLayerNumber
		<< "," << std::setw(1) << fCellNumber
		<< "," << std::setw(1) << fWireNumber << ")";
	TIter next(this);
	U4CDCHit *hp;
	Int_t n = 0;
	while ((hp = (U4CDCHit *)next())) {
		if (n != 0) std::cerr << "                  ";
		std::cerr << std::setw(3) << ++n << " ";
		hp->DebugPrint(opt);
	}
	std::cerr.unsetf(std::ios::right);
	std::cerr.unsetf(std::ios::fixed);
}


void U4CDCHitWire::ChannelToWire(U4CDCCluster *cluster,
  			       Int_t &layer, Int_t &cell, Int_t &wire, 
  			       Double_t &x, Double_t &y)
{
        //U4CDCCluster* cluster = ht->GetClusterPtr();
        
	Int_t cr  = 0;		// We have only 1 crate now.
	Int_t st  = cluster->GetStationNo();
	Int_t ch  = cluster->GetChannelNo();
	Int_t dch, id;
	
	U4CDCRawData     *rd = (U4CDCRawData *)gJSF->FindModule("U4CDCRawData");
	U4CDCBeginRunBuf *br = rd->GetBeginRunBuf();
	Int_t fmtid = br->GetHeader()->GetFormatID();
	switch (fmtid) {
	    case 0: {					// 98 format 
	        dch = ch;
		break;
	    }
	    case 1: {					// new format
	        br->EChannelToDChannel(cr,st,ch,id,dch);
		break;
	    }
	    default: {
	        std::cerr	<< "////// ERROR in U4CDCHitWire::ChannelToWire" << std::endl 
	        	<< "Invalid format ID = " << fmtid << std::endl;
		break;
	    }
	}
	ChannelToWire(dch,layer,cell,wire,x,y);	
}


void U4CDCHitWire::ChannelToWire(Int_t dch,
  			       Int_t &layer, Int_t &cell, Int_t &wire, 
  			       Double_t &x, Double_t &y)
{
//////////////// Very Temporary ////////////////////
/////	Should be moved to data fields  ////////////
/////   The values are only approximate ////////////

	Int_t nwirepercell   = 5;
	Double_t cellheight  = 10.;
	Double_t cellwidth   = 5.;
	Double_t wirespacing = 1.;

/////// This works only for baby chamber ///////////

	Int_t c  = dch/nwirepercell;
//	layer    = c%2;
	layer    = (c+1)/2;
	cell     = c/2;
	wire     = dch%nwirepercell;
//	x        = (1-c)*cellwidth;  
	x        = (c > 1 ? -1 : c)*cellwidth;  
        y        = layer*cellheight + wire*wirespacing;

//////////////// Very Temporary ////////////////////
}



